--[[-------------------------------------------------------------------------------

	Author: thelegendaryof <thelegendaryof@gmail.com>	
	Package: @FastReport
	Version: 1.1.9.1

	Game: The Elder Scrolls - Online
	File: FastReport.lua
	Path: \
	
	Description:
	
	All the code belongs to us :3
	
	Credits:
	
	http://wiki.esoui.com/ - For providing the API Info

	Wykkyd 			- For his buffer snippet and inspiration
	WarLegend		- For code inspiration
	Seerah 			- For LibStub and LibAddonSettings
	
  ]]-------------------------------------------------------------------------------

local FastReport = {}
local FastReport_DW = nil
local FastReport_SCENE = nil
local FastReport_FRAG = nil
local FastReport_SaveData = nil
local FastReport_TempSaveData = nil
local FastReport_API = LibStub:NewLibrary("FastReportAPI-1.0", 1)

local LAM = LibStub:GetLibrary("LibAddonMenu-1.0")
local LibOrangUtils = LibStub("LibOrangUtils-1.0")

local function debug_print(out)
	--d("[FastReport] "..tostring(out))
end

-----------------------------
-- TRACKING
-----------------------------

-- UpdateTrackedPlayer Function
local function UpdateTrackedPlayer(index, name, class, class_id, level, map, pos)	
	if(not index or not name or name == "") then do return false end end

	local player = {}

	player.name = name
	
	if(class and class ~= "") then player.class = class end
	if(class_id and class_ud ~= "") then player.class_id = class_id end
	if(level and level ~= "") then player.level = level end
	if(map and map ~= "") then player.map = map end
	if(pos) then player.pos = pos end
	
	player.time = GetTime() 
	player.index = index
	
	debug_print("UpdateTrackedPlayer - Update player at index:"..tostring(index))
	
	--FastReport.targetList[index] = nil
	FastReport.targetList[index] = player
	
	debug_print("UpdateTrackedPlayer - Updated player with name:"..FastReport.targetList[index].name)
	
	return FastReport.targetList[index]
end

-- DeleteTrackedPlayer Function
local function DeleteTrackedPlayer(index)
	index = index or 1 -- if no index has been given it deletes the first tracked player
	
	if(FastReport.targetList[index] ~= nil) then FastReport.targetList[index] = nil 
	else do return false end
	end
end

-- AddTrackedPlayer Function
local function AddTrackedPlayer(name, class, class_id, level, map, pos)
	local index = #FastReport.targetList + 1
	
	return UpdateTrackedPlayer(index, name, class, class_id, level, map, pos)
end

-- GetTrackedPlayer Function
local function GetTrackedPlayer(index)
	if(not index) then do return false end end
	
	if(FastReport.targetList[index] ~= nil) then return FastReport.targetList[index]
	else do return false end
	end
end

-- ExistsTrackedPlayer Function
local function ExistsTrackedPlayer(name)	
	return FastReport.targetList[name] or false
end

-----------------------------
-- UI
-----------------------------

-- Cleanup Text in the Templates
local function CleanupText(text)
	text = trim(text) -- trim me baby one more time
	text = string.gsub(text, "\t+", "") -- strip tabs 	

	return text
end

-- Load Saves
local function InitSaveData()
	Default	= {
			["HUD_Enabled"] = true,
			["HUD_ShowTarget"] = true,
			["HUD_Align"] = "right",
			["DetailWindow_Enabled"] = true,
			["TEMPLATES"] = FastReport_Templates
	}

	-- Load saved variables
	FastReport_SaveData = ZO_SavedVars:NewAccountWide( "FastReportSaveData", 1, nil, Default, nil )
end

-- Prepare the Report-Templates, remove trailing NL and tabs
local function PrepareReportTemplates()
	for k, v in pairs(FastReport_SaveData["TEMPLATES"]) do 
		FastReport_SaveData["TEMPLATES"][k] = CleanupText(v)
	end
end

local function GetRawReportTemplate(template_name)
	template_name = template_name:upper()
	template_name = LibOrangUtils:str_replace(tostring(template_name), "%s", "_")

	if(FastReport_SaveData and FastReport_SaveData["TEMPLATES"] and FastReport_SaveData["TEMPLATES"][template_name]) then
		local 	text = FastReport_SaveData["TEMPLATES"][template_name]
				text = CleanupText(text)
		return 	text
	end
	
	return false
end

local function GetDefaultReportTemplate(template_name)
	template_name = template_name:upper()
	template_name = LibOrangUtils:str_replace(tostring(template_name), "%s", "_")
	
	return FastReport_Templates[template_name]
end

local function SaveRawReportTemplate(template_name, text)
	template_name = template_name:upper()
	template_name = LibOrangUtils:str_replace(tostring(template_name), "%s", "_")

	if(FastReport_SaveData and FastReport_SaveData["TEMPLATES"]) then
		FastReport_SaveData["TEMPLATES"][template_name] = text
		return true
	end
	
	return false
end

-- Grab the Report-Template
local function GetReportTemplate(template_name, player)
	template_name = template_name or "DEFAULT"
	template_name = template_name:upper()
	template_name = LibOrangUtils:str_replace(tostring(template_name), "%s", "_")
	
	if(player and type(player) == "table") then
		player = player
	else
		player = FastReport.current
	end
	
	player.pos = player.pos or {x=nil, y=nil, z=nil}
	
	-- map n' the template variables correctly
	local map = { 
		name = player.name, 
		class = player.class,
		class_id = player.class_id,
		level = player.level,
		map = player.map,
		x = player.pos.x,
		y = player.pos.y,
		z = player.pos.z,
		time = player.time ~= "" and player.time or GetTime()
	}
	
	debug_print("GetReportTemplate - Template-Name:"..tostring(template_name)..", Player-Name:"..tostring(player.name))
	
	local text = ""
	
	if(FastReport_SaveData and FastReport_SaveData["TEMPLATES"] and FastReport_SaveData["TEMPLATES"][template_name]	) then
		text = GetRawReportTemplate(template_name)
	else
		text = FastReport_Templates[template_name]
	end
	
	text = replace_vars(text, map)
		  
	return text
end


-- Create Settings Panel
local function CreateSettingsPanel()
	local panelID = LAM:CreateControlPanel("FastReport_Settings", "|cFF2222Fast|rReport")
	
	-- HUD
	LAM:AddHeader(panelID, "FastReport_Settings_HUD", "HUD")
	LAM:AddCheckbox(panelID, "FastReport_Settings_HUD_Enable", "Enable HUD", "Show/hide the Target HUD", 
		function() return FastReport_SaveData.HUD_Enabled end,  -- getFunc
		function() 												-- setFunc
			FastReport_SaveData.HUD_Enabled = not FastReport_SaveData.HUD_Enabled
			
			FastReport_HUD:SetHidden(not FastReport_SaveData.HUD_Enabled)
		end
	)
	
	LAM:AddCheckbox(panelID, "FastReport_Settings_HUD_ShowTarget", "Show Target", "Show/hide the targets name, class and level", 
		function() return FastReport_SaveData.HUD_ShowTarget end,  -- getFunc
		function() 												-- setFunc
			FastReport_SaveData.HUD_ShowTarget = not FastReport_SaveData.HUD_ShowTarget
			
			FastReport_HUD_currentTarget:SetHidden(not FastReport_SaveData.HUD_ShowTarget)
			FastReport_HUD_classIcon:SetHidden(not FastReport_SaveData.HUD_ShowTarget)
		end
	)	
	
	-- DETAIL WINDOW
	LAM:AddHeader(panelID, "FastReport_Settings_Detail", "Detail-Window")
	LAM:AddCheckbox(panelID, "FastReport_Settings_Detail_Enable", "Enable Detail-Window", "Show/Hide the Detail Window", 
		function() return FastReport_SaveData.DetailWindow_Enabled end,  -- getFunc
		function() 												-- setFunc
			FastReport_SaveData.DetailWindow_Enabled = not FastReport_SaveData.DetailWindow_Enabled
		end
	)	
end

-- Create Template-Settings Window
local function CreateTemplateSettings()

end

-- Align HUD
local function HUD_SetHorizontalAlign(align)
	local align = align or "right"
	
	if(align == "right") then
		FastReport_HUD_HistoryButton:ClearAnchors()
		FastReport_HUD_currentTarget:ClearAnchors()
		FastReport_HUD_classIcon:ClearAnchors()
		
		FastReport_HUD_HistoryButton:SetAnchor(RIGHT, FastReport_HUD, RIGHT, -16, 0) -- <Anchor point="RIGHT" offsetX="-16" offsetY="0"/>
		FastReport_HUD_currentTarget:SetAnchor(RIGHT, FastReport_HUD, RIGHT, -90, 0) -- <Anchor point="RIGHT" offsetX="-90" offsetY="0" />
		FastReport_HUD_classIcon:SetAnchor(RIGHT, FastReport_HUD, RIGHT, -70, 0) -- <Anchor point="RIGHT" offsetX="-70" offsetY="0" />
		
		if(FastReport_SaveData) then
			FastReport_SaveData.HUD_Align = "right"
		end		
	elseif(align == "left") then
		FastReport_HUD_HistoryButton:ClearAnchors()
		FastReport_HUD_currentTarget:ClearAnchors()
		FastReport_HUD_classIcon:ClearAnchors()
									
		FastReport_HUD_HistoryButton:SetAnchor(LEFT, FastReport_HUD, LEFT, 16, 0)
		FastReport_HUD_currentTarget:SetAnchor(LEFT, FastReport_HUD, LEFT, 90, 0)
		FastReport_HUD_classIcon:SetAnchor(LEFT, FastReport_HUD, LEFT, 70, 0)

		if(FastReport_SaveData) then
			FastReport_SaveData.HUD_Align = "left"
		end
	end
end

-- Init HUD
local function InitHUD()
	FastReport_HUD:SetHidden(not FastReport_SaveData.HUD_Enabled)
	
	HUD_SetHorizontalAlign(FastReport_SaveData.HUD_Align)
	
	if(FastReport_SaveData.HUD_Position and type(FastReport_SaveData.HUD_Position) == "table") then
		FastReport_HUD:ClearAnchors()
		FastReport_HUD:SetAnchor(FastReport_SaveData.HUD_Position.a, GuiRoot, FastReport_SaveData.HUD_Position.b, FastReport_SaveData.HUD_Position.x, FastReport_SaveData.HUD_Position.y)
	end
	
	FastReport_HUD:SetHandler("OnMouseDown", function()
		FastReport.HUD_IsMoving = true
	end)
	
	FastReport_HUD:SetHandler("OnMouseUp", function()
		--FastReport_HUD:StopMovingOrResizing()
		FastReport.HUD_IsMoving = false
		
		local _,a,_,b,x,y = FastReport_HUD:GetAnchor()
		FastReport_SaveData.HUD_Position = {["a"]=a, ["b"]=b, ["x"]=x, ["y"]=y}

	end)
	
	FastReport_HUD:SetHandler("OnUpdate", function()
		if(FastReport.HUD_IsMoving and FastReport.HUD_IsMoving == true) then
			if not FastReport_BufferReached("HUD_OnUpdate", 0.33) then do return end end

					
			local _,a,_,b,x,y = FastReport_HUD:GetAnchor()
			local g_h_c = GuiRoot:GetWidth()*0.5
			
			if(a == 12 or a == 9) then x = x + GuiRoot:GetWidth() end		
			
			if(x < g_h_c) then 
				HUD_SetHorizontalAlign("left")
			else
				HUD_SetHorizontalAlign("right")
			end
			
			--d("OnUpdate -  x:"..tostring(x)..", ghc:"..tostring(g_h_c)..", align: "..tostring(FastReport_SaveData.HUD_Align))
			--d("a: "..tostring(a))
		
		end
	end)
end

-- Create FastReport Detail-Window and attach it to the Support Scene
local function CreateDetailWindow()
	--WINDOW_MANAGER:CreateTopLevelWindow("FastReport_VC")
	--FastReport_DW = WINDOW_MANAGER:CreateTopLevelWindow("FastReport_DW", FastReport_VC)
	FastReport_DW = WINDOW_MANAGER:CreateTopLevelWindow("FastReport_DW")
	FastReport_SCENE = SCENE_MANAGER.scenes.helpCustomerSupport
	FastReport_FRAG = ZO_FadeSceneFragment:New(FastReport_DW)
	
	FastReport_SCENE:AddFragment(FastReport_FRAG)
	
	-- TODO: Hm, maybe ... ah nevermind.
	
	FastReport_SCENE:RegisterCallback("StateChange", function(oldState, newState) 
		if(newState == "hiding") then -- and FastReport_SaveData.DetailWindow_Enabled == true) then
			--FastReport.tiggeredFromWithin = false
			--FastReport_DW:SetHidden(true)
			ZO_Menus:SetTopmost(false)
			FastReport_DisableEditMode()
		elseif(newState == "showing") then -- and FastReport.tiggeredFromWithin == true  and FastReport_SaveData.DetailWindow_Enabled == true) then
			--FastReport_DW:SetHidden(false)
			ZO_Menus:SetTopmost(true)
		end
	end)
	
	PrepareReportTemplates()
	
	local GuiRoot = WINDOW_MANAGER:GetControlByName("GuiRoot")
	local w = (GuiRoot:GetWidth()/2)*0.6
	local h = GuiRoot:GetHeight()

	FastReport_DW:SetHidden(true)
	--FastReport_DW:SetDimensions(w, h)
	FastReport_DW:SetAnchor(LEFT, GuiRoot, LEFT, 0, 0)
	--FastReport_DW:SetAnchor(CENTER, GuiRoot, CENTER, 0, 0)
	FastReport_DW:SetDimensions(w, h)
	
	FastReport_DW:SetTopmost(true)
	--FastReport_DW:SetDrawLayer(3)
	--FastReport_DW:SetDrawTier(3)
	--FastReport_DW:SetDrawLevel(3)
	FastReport_DW:SetInheritAlpha(true)
	FastReport_DW:SetInheritScale(true)
	
	-- BG
	WINDOW_MANAGER:CreateControlFromVirtual("FastReport_DW_BG", FastReport_DW, "ZO_RightFootPrintBackground")
	
	FastReport_DW_BG:SetAlpha(1)
	--FastReport_DW_BG:SetDrawLayer(2)
	
	local bg_w = FastReport_DW_BG:GetWidth()
	local bg_h = FastReport_DW_BG:GetHeight()	
	
	FastReport_DW_BG:ClearAnchors()
	FastReport_DW_BG:SetAnchor(LEFT, FastReport_DW, LEFT, 0, (bg_h*0.035))
	
	-- CLOSE
	WINDOW_MANAGER:CreateControlFromVirtual("FastReport_DW_Close", FastReport_DW, "ZO_CloseButton")
	FastReport_DW_Close:SetAnchor(TOPRIGHT, FastReport_DW_BG, TOPRIGHT, -10, 8)
	--FastReport_DW_Close:SetDrawLayer(3)
	FastReport_DW_Close:SetHandler("OnClicked", function() 
		SCENE_MANAGER:SetInUIMode() 
		FastReport_DW:SetHidden(true)  
		FastReport_DisableEditMode()
	end)
	
	-- TITLEBAR
	WINDOW_MANAGER:CreateControlFromVirtual("FastReport_DW_Title", FastReport_DW, "ZO_PanelTitle")
	--FastReport_DW_Title:SetDimensions(w,50)
	FastReport_DW_Title:ClearAnchors()
	FastReport_DW_Title:SetAnchor(TOPLEFT, FastReport_DW_BG, TOPLEFT, 0, (-bg_h*0.01))
	--FastReport_DW_Title:SetDrawLayer(3)
	FastReport_DW_TitleLabel:SetText("|cFF2222Fast|rReport")
	--FastReport_DW_TitleLabel:SetColor(230/255,231/255,175/255, 1)
	FastReport_DW_TitleDivider:ClearAnchors()
	FastReport_DW_TitleDivider:SetAnchor(TOP,FastReport_DW_Title, BOTTOM, (-bg_w*0.05), 0) 
	FastReport_DW_TitleDivider:SetDimensions(bg_w,4)
	
	--[[
	<Control name="$(parent)Dropdown" inherits="ZO_ComboBox">
	<Dimensions x="300" y="28" />
	<Anchor point="RIGHT" offsetX="-35" />
	</Control>
	--]]
	
	-- TEMPLATE DROPDOWN
	WINDOW_MANAGER:CreateControlFromVirtual("FastReport_DW_PlayerDetails_TEMPLATE", FastReport_DW, "ZO_ComboBox")
	FastReport_DW_PlayerDetails_TEMPLATE:SetAnchor(TOPLEFT, FastReport_DW_Title, TOPLEFT, (bg_w*0.053), 80)
	FastReport_DW_PlayerDetails_TEMPLATE:SetDimensions((bg_w*0.9), 40)
	--FastReport_DW_PlayerDetails_TEMPLATE:SetDrawLayer(3)
	FastReport_DW_PlayerDetails_TEMPLATESelectedItemText:SetColor(197/255, 194/255, 158/255, 1)
	FastReport_DW_PlayerDetails_TEMPLATESelectedItemText:SetFont("ZoFontHeader")
	
	-- PLAYERDETAILS BG
	WINDOW_MANAGER:CreateControlFromVirtual("FastReport_DW_PlayerDetails_BG", FastReport_DW, "ZO_DefaultBackdrop")
	--FastReport_DW_PlayerDetails_BG:SetAlpha(0.85)
	FastReport_DW_PlayerDetails_BG:ClearAnchors()
	FastReport_DW_PlayerDetails_BG:SetAnchor(TOPLEFT, FastReport_DW_PlayerDetails_TEMPLATE, TOPLEFT, 0, 50)
	FastReport_DW_PlayerDetails_BG:SetDimensions((bg_w*0.9), ((bg_h*0.33)))
	FastReport_DW_PlayerDetails_BG:SetDrawLayer(1)
	--FastReport_DW_PlayerDetails_BG:SetHidden(true)
	FastReport_DW_PlayerDetails_BG:SetEdgeColor(0,0,0,1)
	--FastReport_DW_PlayerDetails_BG:SetInsets(0,0,0,0)
	--FastReport_DW_PlayerDetails_BG:SetEdgeTexture("EsoUI/Art/Tooltips/munge_overlay.dds", 128, 16)
	
	-- PLAYERDETAILS EDITBOX
	WINDOW_MANAGER:CreateControlFromVirtual("FastReport_DW_PlayerDetails_EditBox", FastReport_DW, "ZO_DefaultEditMultiLineForBackdrop")
	FastReport_DW_PlayerDetails_EditBox:SetAnchor(TOPLEFT, FastReport_DW_PlayerDetails_BG, TOPLEFT, 10, 7)
	FastReport_DW_PlayerDetails_EditBox:SetAnchor(BOTTOMRIGHT, FastReport_DW_PlayerDetails_BG, BOTTOMRIGHT, -10, -7)
	FastReport_DW_PlayerDetails_EditBox:SetMaxInputChars(-1)
	FastReport_DW_PlayerDetails_EditBox:SetFont("ZoFontEdit")
	FastReport_DW_PlayerDetails_EditBox:SetEditEnabled(false)
	--FastReport_DW_PlayerDetails_EditBox:SetKeyboardEnabled(false)
	FastReport_DW_PlayerDetails_EditBox:SetDrawLayer(3)
	FastReport_DW_PlayerDetails_EditBox:SetColor(207/255, 220/255, 189/255, 1)
	--FastReport_DW_PlayerDetails_EditBox:SetHidden(true)
	
	-- Well, set it once.
	FastReport_DW_PlayerDetails_EditBox:SetText(GetReportTemplate())

	-- 3-Button Jacket
	for i = 1, 3, 1 do
		WINDOW_MANAGER:CreateControlFromVirtual("FastReport_DW_Button"..tostring(i), FastReport_DW, "ZO_DefaultButton")
		
		local button = _G["FastReport_DW_Button"..tostring(i)]
		
		button:SetAnchor(TOPRIGHT, FastReport_DW_PlayerDetails_BG, BOTTOMRIGHT, 5+(-165*(i-1)), 11)
		button:SetDimensions(160, 32)
		button:SetHidden(true)
		button:SetMouseEnabled(true)
		button:SetText("Button")
		button:SetFont("ZoFontHeader")
		--button:SetNormalTexture(nil)
		button:SetMouseOverTexture(nil)
		button:SetPressedTexture(nil)
		button:SetDisabledTexture(nil)
		--button:SetHorizontalAlignment(TEXT_ALIGN_RIGHT)
		--button:SetVerticalAlignment(TEXT_ALIGN_TOP)
		
		--button:SetDrawLayer(3)
		--button:SetDrawLevel(3)
	end
	
	-- HISTORY TITLE
	WINDOW_MANAGER:CreateControlFromVirtual("FastReport_History_Title", FastReport_DW, "ZO_PanelTitle")
	--FastReport_History_Title:SetDimensions(w, 50)
	FastReport_History_Title:ClearAnchors()
	FastReport_History_Title:SetAnchor(TOPLEFT, FastReport_DW_PlayerDetails_BG, BOTTOMLEFT, -45, 40)
	--FastReport_History_Title:SetDrawLayer(3)
	FastReport_History_TitleLabel:SetText("|cFF2222PLAYER|rHISTORY")
	--FastReport_History_TitleLabel:SetColor(230/255,231/255,175/255, 1)
	FastReport_History_TitleDivider:ClearAnchors()
	FastReport_History_TitleDivider:SetAnchor(TOP,FastReport_History_Title, BOTTOM, (-bg_w*0.05), 0) 
	FastReport_History_TitleDivider:SetDimensions(bg_w,4)
	
	-- HISTORY WINDOW
	FastReport_History:SetParent(FastReport_DW)
	--FastReport_History:SetDrawLayer(3)
	FastReport_History:ClearAnchors()
	FastReport_History:SetAnchor(TOPLEFT, FastReport_History_Title, BOTTOMLEFT, (bg_w*0.06), 10)
	FastReport_History:SetAnchor(BOTTOMRIGHT, FastReport_DW_BG, BOTTOMRIGHT, -(bg_w*0.06), -60)
	FastReport_History:SetHidden(false)
	
	FastReport_History_Headers:ClearAnchors()
	FastReport_History_Headers:SetAnchor(TOPLEFT)
	FastReport_History_Headers:SetAnchor(TOPRIGHT)
	
	local base_w = (bg_w/6)-15 
	local id_w	 = base_w * 0.5
	local name_w = base_w * 1.25
	local location_w = base_w * 1.25
	local class_w = base_w * 1
	local lvl_w = base_w * 1
	local time_w = base_w * 1
	
	FastReport_History_Headers_ID:SetWidth(id_w)
	FastReport_History_Headers_Name:SetWidth(name_w)
	FastReport_History_Headers_Zone:SetWidth(location_w)
	FastReport_History_Headers_Class:SetWidth(class_w)
	FastReport_History_Headers_Level:SetWidth(lvl_w)
	FastReport_History_Headers_Time:SetWidth(time_w)

	-- TRACKLIST
	--[[
	WINDOW_MANAGER:CreateControlFromVirtual("FastReport_DW_TrackList", FastReport_DW, "ZO_ScrollContainer")
	FastReport_DW_TrackList:SetAnchor(TOPLEFT, FastReport_DW_PlayerDetails, TOPLEFT)
	FastReport_DW_TrackList:SetAnchor(BOTTOMRIGHT, FastReport_DW, BOTTOMRIGHT)
	--FastReport_DW_List
	  ]]--
end

local function RedrawDetailWindow()
	local GuiRoot = WINDOW_MANAGER:GetControlByName("GuiRoot")
	local w = (GuiRoot:GetWidth()/2)*0.6
	local h = GuiRoot:GetHeight()
	
	--FastReport_DW:SetDimensions(w, h)	
	
	FastReport_DW_BG:ClearAnchors()
	FastReport_DW_BG:SetAnchor(LEFT, FastReport_DW, LEFT, 0, (bg_h*0.035))	
	
	local bg_w = FastReport_DW_BG:GetWidth()
	local bg_h = FastReport_DW_BG:GetHeight()		
	
	FastReport_DW_Title:ClearAnchors()
	FastReport_DW_Title:SetAnchor(TOPLEFT, FastReport_DW_BG, TOPLEFT, 0, (-bg_h*0.01))
	FastReport_DW_TitleDivider:ClearAnchors()
	FastReport_DW_TitleDivider:SetAnchor(TOP,FastReport_DW_Title, BOTTOM, (-bg_w*0.05), 0) 
	FastReport_DW_TitleDivider:SetDimensions(bg_w,4)	
	
	FastReport_DW_PlayerDetails_TEMPLATE:ClearAnchors()
	FastReport_DW_PlayerDetails_TEMPLATE:SetAnchor(TOPLEFT, FastReport_DW_Title, TOPLEFT, (bg_w*0.053), 80)
	FastReport_DW_PlayerDetails_TEMPLATE:SetDimensions((bg_w*0.9), 40)	
	
	FastReport_DW_PlayerDetails_BG:ClearAnchors()
	FastReport_DW_PlayerDetails_BG:SetAnchor(TOPLEFT, FastReport_DW_PlayerDetails_TEMPLATE, TOPLEFT, 0, 50)
	FastReport_DW_PlayerDetails_BG:SetDimensions((bg_w*0.9), ((bg_h*0.33)))	
	
	FastReport_History_TitleDivider:ClearAnchors()
	FastReport_History_TitleDivider:SetAnchor(TOP,FastReport_History_Title, BOTTOM, (-bg_w*0.05), 0) 
	FastReport_History_TitleDivider:SetDimensions(bg_w,4)

	FastReport_History:ClearAnchors()
	FastReport_History:SetAnchor(TOPLEFT, FastReport_History_Title, BOTTOMLEFT, (bg_w*0.06), 10)
	FastReport_History:SetAnchor(BOTTOMRIGHT, FastReport_DW_BG, BOTTOMRIGHT, -(bg_w*0.06), -60)	
	
	local base_w = (bg_w/6)-15 
	local id_w	 = base_w * 0.5
	local name_w = base_w * 1.25
	local location_w = base_w * 1.25
	local class_w = base_w * 1
	local lvl_w = base_w * 1
	local time_w = base_w * 1
	
	FastReport_History_Headers_ID:SetWidth(id_w)
	FastReport_History_Headers_Name:SetWidth(name_w)
	FastReport_History_Headers_Zone:SetWidth(location_w)
	FastReport_History_Headers_Class:SetWidth(class_w)
	FastReport_History_Headers_Level:SetWidth(lvl_w)
	FastReport_History_Headers_Time:SetWidth(time_w)	

end

-- Update Report Template for Tracked-Player
local function SetReportTemplate_Tracked(index)
	local player = GetTrackedPlayer(index)
	if(not index or index == "" or not player or player == false) then  
		-- create empty player 
		player = { index = "0", name = "", class = "", class_id = 0, level = 0, map = "", pos = {x=0,y=0,z=0}, time = GetTime() }	
	end

	local text = GetReportTemplate("TRACKED", player)
	
	debug_print("SetReportTemplate - Index:"..tostring(index)..", Player-Name: "..tostring(player.name))
	
	if(text and text ~= "") then
		FastReport_DW_PlayerDetails_EditBox:SetText(text)	  
	end
end

-- Update Report Template for Default (/report or no Tracked-Player found)
local function SetReportTemplate_Default(name)
	local text = ""
	
	if(not name or name == "") then 
		text = GetReportTemplate()		
	else
		local player = {
			name = name,
			time = GetTime()
		}
		
		text = GetReportTemplate("DEFAULT", player)		
	end
	
	debug_print("SetReportTemplate_Default - Player-Name:"..tostring(name))
	
	if(text and text ~= "") then
		FastReport_DW_PlayerDetails_EditBox:SetText(text)	  
	end
end

-- SetReportTemplate for addons / anything else
local function SetReportTemplate(template_name, name)
	if(not template_name or template_name == "") then do return nil end end
	
	name = name or ""
	
	--d("SetReportTemplate()")

	template_name = template_name:upper()
	template_name = LibOrangUtils:str_replace(tostring(template_name), "%s", "_")

	local map		= GetPlayerLocationName()
	local x,y,z 	= GetMapPlayerPosition("player")
	if(x) then x 	= LibOrangUtils:round(x, 5) end
	if(y) then y 	= LibOrangUtils:round(y, 5) end
	if(z) then z 	= LibOrangUtils:round(z, 5) end
	local pos 		= { x = x, y = y, z = z }
	
	local player = {
		name = name,
		time = GetTime(),
		map = map,
		pos = pos
	}
	
	text = GetReportTemplate(template_name, player)		
	
	debug_print("SetReportTemplate "..tostring(template_name).." - Player-Name:"..tostring(name))
	
	if(text and text ~= "") then
		FastReport_DW_PlayerDetails_EditBox:SetText(text)	  
	end	
end

local function DropdownSetRawTemplate(template_name)	
	if(template_name) then 
		template_name = template_name:upper()
		template_name = LibOrangUtils:str_replace(tostring(template_name), "%s", "_")
		
		if(FastReport_SaveData["TEMPLATES"] and FastReport_SaveData["TEMPLATES"][template_name]) then 
			--FastReport_SaveData["TEMPLATES"][template_name] = GetDefaultReportTemplate(template_name) 
			local text = GetRawReportTemplate(template_name)
			
			if(text) then FastReport_DW_PlayerDetails_EditBox:SetText(text) end
		end
	end
end

local function DropdownSetTemplate(template_name)
	template_name = template_name or "DEFAULT"
	template_name = template_name:upper()
	template_name = LibOrangUtils:str_replace(tostring(template_name), "%s", "_")
		
	if(template_name == "DEFAULT") then
		SetReportTemplate_Default()
	elseif(template_name == "TRACKED") then
		SetReportTemplate_Tracked(FastReport.current.index)
	else
		SetReportTemplate(template_name)
	end
end

local function DropdownOnItemSelect(self, choiceText, choice)
	--d("itemselect")
	local args = nil
	if(choiceText and LibOrangUtils and LibOrangUtils.str_split) then args = LibOrangUtils:str_split(choiceText, "%s%-%s") end
			
	if(args and type(args) == "table") then 
		local 	template_name = args[1]
				template_name = template_name:upper()
				template_name = LibOrangUtils:str_replace(tostring(template_name), "%s", "_")
		
		-- args found
		--d("itemselect -> "..tostring(template_name)) 
		if(FastReport.editmode == true) then
			DropdownSetRawTemplate(template_name)
		else 
			DropdownSetTemplate(template_name)
		end
	else
		-- choiceText
	end
end

local function DropdownAddItem(name)
	if(not name or name == "") then do return false end end

	local dropdown = FastReport_DW_PlayerDetails_TEMPLATE.m_comboBox
	local entry = dropdown:CreateItemEntry(name, DropdownOnItemSelect) 
	dropdown:AddItem(entry)
end

local function InitTemplateDropdown()
	local dropdown = FastReport_DW_PlayerDetails_TEMPLATE.m_comboBox
	local items = { 
		'DEFAULT - USED FOR GENERAL REPORTS', 
		'TRACKED - USED FOR TRACKED PLAYERS' 
	} 
	
	for i=1,#items do
		DropdownAddItem(items[i])
	end
	
	dropdown:SelectFirstItem()
end

local function DropdownSelect(name)
	local dropdown = FastReport_DW_PlayerDetails_TEMPLATE.m_comboBox
	local items = dropdown:GetItems()
	
	for i, item in ipairs(items) do
		debug_print("Checking "..tostring(item["name"]).." - for: "..name)
		if(type(item) == "table" and item["name"] and item["name"]:find(name)) then dropdown:SetSelectedItem(item["name"]) return true end
	end
	
	return false
end

local function DropdownGetSelected() 
	local selected = FastReport_DW_PlayerDetails_TEMPLATESelectedItemText:GetText()
	local dropdown = FastReport_DW_PlayerDetails_TEMPLATE.m_comboBox	
	local items = dropdown:GetItems()
		
	for i, item in ipairs(items) do
		debug_print("Checking "..tostring(item["name"]).." - for: "..selected)
		if(type(item) == "table" and item["name"]) then 
			if(item["name"] == selected or item["name"]:find(selected)) then return item end
		end
	end
	
	return false
end

local function DropdownGetSelectedTemplate() 
	local selected = DropdownGetSelected()
	local args = nil
	
	--d("DropdownGetSelectedTemplate")	
	
	if(selected and selected["name"] and LibOrangUtils and LibOrangUtils.str_split) then
		--d("ARK")
		args = LibOrangUtils:str_split(selected["name"], "%s%-%s")
		if(args and type(args) == "table") then 
			--d("SELECTED: "..args[1])
			local template_name = args[1]
			
			return args[1]
		end
	end
	
	return false
end

local function AddonReportTemplate(name, description, text)
	if(not name or not description or not text) then do return nil end end

	description = description:upper()
	
	local template_name = LibOrangUtils:str_replace(tostring(name), "%s", "_")
	local dropdown_item = tostring(name).." - "..tostring(description)
	local dropdown = FastReport_DW_PlayerDetails_TEMPLATE.m_comboBox	
	local items = dropdown:GetItems()
	local exists = false
	
	--d("Adding: "..dropdown_item)
		
	for i, item in ipairs(items) do
		if(item and type(item) == "table" and (item["name"] == name or item["name"]:find(name))) then exists = true end
	end
	
	if(exists == false) then
		--d("yay doesn't exist yet!")
		SaveRawReportTemplate(template_name, text)
		DropdownAddItem(dropdown_item)
		DropdownSelect(name)
	else
		--d("bah - exists updating ...")
		SaveRawReportTemplate(template_name, text)
		SetReportTemplate(template_name)
		DropdownSelect(name)
	end
end

function FastReport_EnableEditMode()
	FastReport.editmode = true

	FastReport_History:SetMouseEnabled(false)
	FastReport_History:SetAlpha(0.3)
	FastReport_History_Title:SetAlpha(0.3)
	FastReport_DW_PlayerDetails_EditBox:SetEditEnabled(true)
	--FastReport_DW_PlayerDetails_EditBox:SetKeyboardEnabled(true)
	
	FastReport_DW_PlayerDetails_BG:SetEdgeColor(2,2,2,1)
	
	FastReport_DW_Button1:SetText("EXIT EDITMODE")	
	FastReport_DW_Button1:SetHandler("OnClicked", function(self) 
		FastReport_DisableEditMode()
	end)	
	
	FastReport_DW_Button2:SetHidden(false)
	FastReport_DW_Button3:SetHidden(false)
	
	local template_name = DropdownGetSelectedTemplate()
	DropdownSetRawTemplate(template_name)	
end

function FastReport_DisableEditMode()
	FastReport.editmode = false

	FastReport_History:SetMouseEnabled(true)
	FastReport_History:SetAlpha(1)
	FastReport_History_Title:SetAlpha(1)
	FastReport_DW_PlayerDetails_EditBox:SetEditEnabled(false)	
	--FastReport_DW_PlayerDetails_EditBox:SetKeyboardEnabled(false)	
	FastReport_DW_PlayerDetails_BG:SetEdgeColor(0.5,0.5,0.5,1)
	FastReport_DW_Button1:SetText("EDIT TEMPLATE")
	FastReport_DW_Button1:SetHidden(false)
	FastReport_DW_Button2:SetHidden(true)
	FastReport_DW_Button3:SetHidden(true)	
	FastReport_DW_Button1:SetHandler("OnClicked", function(self) 
		FastReport_EnableEditMode()
	end)	
	
	local template_name = DropdownGetSelectedTemplate()
	DropdownSetTemplate(template_name)		
end

local function InitEditMode()
	FastReport_DW_Button1:SetText("EDIT TEMPLATE")
	
	FastReport_DW_Button2:SetText("SAVE TEMPLATE")	
	FastReport_DW_Button2:SetHandler("OnClicked", function() 
		local template_name = DropdownGetSelectedTemplate()
		if(template_name) then 
			text = FastReport_DW_PlayerDetails_EditBox:GetText()
			if(text) then SaveRawReportTemplate(template_name, text) end
		end
	end)
	
	FastReport_DW_Button3:SetText("LOAD DEFAULT")	
	FastReport_DW_Button3:SetHandler("OnClicked", function() 
		local template_name = DropdownGetSelectedTemplate()
		local default_text = GetDefaultReportTemplate(template_name)
		
		if(default_text) then 
			SaveRawReportTemplate(template_name, default_text) 
			if(FastReport.editmode == true) then
				DropdownSetRawTemplate(template_name)
			else 
				DropdownSetTemplate(template_name1)
			end
		end
	end)
	
	FastReport_DisableEditMode()	
end

-----------------------------
-- REPORTING
-----------------------------

local function InternalReportPlayer(name, type)
	FastReport.tiggeredFromWithin = true

	FastReport_DisableEditMode()	

	if((name and name ~= "" and not type) or type == "DEFAULT") then
		DropdownSelect("DEFAULT")
		SetReportTemplate_Default(name)
		ZO_FEEDBACK:ReportPlayer(name)
	elseif((not name or name == "" and FastReport.current.name and not type) or type == "TARGET") then
		DropdownSelect("TRACKED")
		SetReportTemplate_Tracked(FastReport.current.index)
		ZO_FEEDBACK:ReportPlayer(FastReport.current.name) 	
	else
		DropdownSelect(type)
		SetReportTemplate(type)
		ZO_FEEDBACK:ReportPlayer(name)
	end
end

-- Report Function
function FastReport_ReportPlayer(name, type)
	--d("ReportPlayer - Name: "..tostring(name))
	InternalReportPlayer(name, type)
end

-- SetTarget Function
local function SetTarget(name, class, class_id, level, map, pos)
	-- class id:
	-- 1: Dragonknight
	-- 2: Sorcerer
	-- 3: Nightblade
	-- 6: Templar
	
	if(class_id) then
		if(class_id == 1) then 
			FastReport_HUD_classIcon:SetTexture("ESOUI/art/contacts/social_classicon_dragonknight.dds") 
			FastReport_HUD_classIcon:SetHidden(false)
		elseif(class_id == 2) then 
			FastReport_HUD_classIcon:SetTexture("ESOUI/art/contacts/social_classicon_sorcerer.dds") 
			FastReport_HUD_classIcon:SetHidden(false)
		elseif(class_id == 3) then 
			FastReport_HUD_classIcon:SetTexture("ESOUI/art/contacts/social_classicon_nightblade.dds")
			FastReport_HUD_classIcon:SetHidden(false)			
		elseif(class_id == 6) then 
			FastReport_HUD_classIcon:SetTexture("ESOUI/art/contacts/social_classicon_templar.dds")
			FastReport_HUD_classIcon:SetHidden(false)			
		else
			FastReport_HUD_classIcon:SetHidden(true)
		end
	end

	debug_print("SetTarget - Player-Name:"..name)
	
	local text = name.." LV"..tostring(level)

	FastReport_HUD_currentTarget:SetText(text)
	
	local exists = ExistsTrackedPlayer(name)
	
	debug_print("SetTarget - Exists Player in Tracking Record?", exists)
	
	if(exists) then
		--d("PLAYER ALREADY EXISTS! UPDATING")
		--Update Player
		FastReport.current = UpdateTrackedPlayer(exists, name, class, class_id, level, map, pos)
	else
		--Insert Player
		FastReport.current = AddTrackedPlayer(name, class, class_id, level, map, pos)
	end
	
end

-----------------------------
-- API
-----------------------------

function FastReport_API:Report(players, text, addon_name, addon_description)

	AddonReportTemplate(addon_name, addon_description, text)

	local template_name = LibOrangUtils:str_replace(tostring(addon_name), "%s", "_")
	
	FastReport_ReportPlayer(players, template_name)
	--if(FastReport_DW_PlayerDetails_EditBox) then FastReport_DW_PlayerDetails_EditBox:SetText(text) end
	--ZO_FEEDBACK:ReportPlayer(players)
end

-----------------------------
-- EVENTS
-----------------------------

-- OnUpdate called every Frame in Dummy.xml
function FastReport_OnUpdate() 
	if not FastReport_BufferReached("CORE_OnUpdate", 0.33) then do return end end
	
	local player = IsUnitPlayer("reticleover")

	--debug_print("OnUpdate - Is Player? "..tostring(player))
	
	if not player == true then do return end end
	
	local name 	= GetUnitName("reticleover")
	
	-- Only Query Detailed Infos if it 's a New Target
	if(name ~= FastReport.current.name and name ~= "" and name) then 
		local class 	= GetUnitClass("reticleover")
		local class_id	= GetUnitClassId("reticleover")	
		local level 	= GetUnitLevel("reticleover")
		local vr		= GetUnitVeteranRank("reticleover")
		local map		= GetPlayerLocationName()
		local x,y,z 	= GetMapPlayerPosition("player")
		if(x) then x 	= LibOrangUtils:round(x, 5) end
		if(y) then y 	= LibOrangUtils:round(y, 5) end
		if(z) then z 	= LibOrangUtils:round(z, 5) end
		local pos 		= { x = x, y = y, z = z }
	
		if(vr and vr > 0) then level = tostring(vr).."VR" end
	
		SetTarget(name, class, class_id, level, map, pos)
	end
	
	--FastReport_HUD:SetHidden(true)
	
end

-- OnCommand called when you type /report
local function FastReport_OnCommand(target)
	FastReport_ReportPlayer(target, "DEFAULT")
	
	debug_print("OnCommand - Target: "..tostring(target))
end

-- OnAddOnLoaded event
local function FastReport_OnAddonLoaded(event, addon)
	if addon == "FastReport" then
		debug_print("OnAddonLoaded - Initializing Addon ...")
	
		SLASH_COMMANDS["/report"] = FastReport_OnCommand
		ZO_CreateStringId("SI_BINDING_NAME_FASTREPORT", "Report Last Targeted Player")
		
		FastReport.editmode = false
		FastReport.current = { name = "", class = "", class_id = 0, level = 0, map = "", pos = {x=0,y=0,z=0}, time = "" }
		FastReport.tiggeredFromWithin = false
		
		-- enable querying for both the key as well as the player-name
		FastReport.targetList = setmetatable({}, {
		  __newindex = function(self,k,v)
			rawset(self,k,v)
			rawset(self,v.name,k)
		  end
		})
		
		FastReport.HUD_IsMoving = false
		
		InitSaveData()
		InitHUD()	
		CreateSettingsPanel()
		CreateDetailWindow()
		InitTemplateDropdown()
		InitEditMode()
		
		debug_print("OnAddonLoaded - Done ...")
		--CreateFastReportDetailWindow()
	end
end

-- OnPlayerActivated event
local function FastReport_OnPlayerActivated()
	FastReport_HUD:SetDrawLevel(0)
	FastReport_HUD:SetDrawLayer(0)
	FastReport_HUD:SetDrawTier(0)
end

EVENT_MANAGER:RegisterForEvent("FastReport_EVENT_ADD_ON_LOADED", EVENT_ADD_ON_LOADED , FastReport_OnAddonLoaded)
EVENT_MANAGER:RegisterForEvent("FastReport_EVENT_PLAYER_ACTIVATED", EVENT_PLAYER_ACTIVATED, FastReport_OnPlayerActivated)